/***************************************************************************
                          hc-mmas-ubqp.cpp  -  description
                             -------------------
    begin                : Fri Mar 21 15:46:28 CET 2003
    copyright            : (C) 2003 by Christian Blum
    email                : cblum@ulb.ac.be
 ***************************************************************************/

/***************************************************************************
    Program's name: hc-mmas-ubqp

    Ant Colony Optimization algorithm to tackle 
    Unary Binary Quadratic Programming

    Copyright (C) 2003  Christian Blum

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Author's contact details: 

    email: cblum@ulb.ac.be
    mail address: Universite Libre de Bruxelles, IRIDIA,
                  Av. Franklin Roosevelt 50, CP 194/6,
                  B-1050 Brussels, Belgium

 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "Timer.h"
#include "Unconstrained_Ant.h"
#include "Random.h"
#include "utilstuff.h"
#include "Solution.h"
#include <string>
#include <list>

// the following five variables are involved in termination criteria issues
int n_of_iter = 1000;
double time_limit = 100.0;
bool time_limit_given = false;
bool iter_limit_given = false;
bool input_file_given = false;

// problem_size: number of variables of the problem instance
int problem_size;

// n_of_ants: the number of ants
long int n_of_ants = 10;

// l_rate: the learning rate (used for updating the pheromone values)
double l_rate = 0.05;

// tau_min, tau_max: lower, respecively upper, limit for the pheromone values
double tau_min = 0.001;
double tau_max = 0.999;

// n_of_trials: the number of trials that is to be executed for the given problem instance
int n_of_trials = 1;

// ls: usage of local search
bool ls = true;

// ants: the list of ants
list<Unconstrained_Ant*> ants;

// q_matrix: the matrix that specifies the problem instance (as read from the instance file)
double** q_matrix;

// c_const: we add c_const to every objective function value in order to have all positive objective function values
double c_const;

// pheromone: the structure that holds the pheromone values
map<int,pair<double,double> >* pheromone;

// t: a variable that is involved in initializing the random generator
time_t t;

// rnd: a random generator
Random* rnd;


/* The method readFile is used to read the problem specification from a file */

void readFile(string fileName) {

  FILE* inputFile = fopen(fileName.c_str(), "r");

  if (fscanf(inputFile, "%ld", &problem_size) < 0) {
    printf("error reading problem size.\n");
    exit(1);
  }  

  q_matrix = DoubleMatrixAlloc(problem_size,problem_size);
  for (int i = 0; i < problem_size; i++) {
    for (int j = 0; j < problem_size; j++) {
      q_matrix[i][j] = 0.0;
    }
  }

  int entries;
  if (fscanf(inputFile, "%ld", &entries) < 0) {
    printf("error reading number of entries.\n");
    exit(1);
  }   
 
  if (fscanf(inputFile, "%lf", &c_const) < 0) {
    printf("error reading c_const.\n");
    exit(1);
  }    

  for (int i = 0; i < entries; i++) {
    int pos1;
    int pos2;
    double val;
    if (fscanf(inputFile, "%ld", &pos1) < 0) {
      printf("error reading first position.\n");
      exit(1);
    }
    if (fscanf(inputFile, "%ld", &pos2) < 0) {
      printf("error reading second position.\n");
      exit(1);
    }
    if (fscanf(inputFile, "%lf", &val) < 0) {
      printf("error reading value.\n");
      exit(1);
    }
    q_matrix[pos1][pos2] = val;
    q_matrix[pos2][pos1] = val;
  }
}

/* The method read_parameters is used to read-in the command line parameters */

void read_parameters(int argc, char **argv)
{
  int iarg=1;

  while (iarg < argc)
  {
    if (strcmp(argv[iarg],"-i")==0) {
      readFile(argv[++iarg]);
      input_file_given = true;
    }
    else if (strcmp(argv[iarg],"-t")==0) { 
      time_limit=atof(argv[++iarg]);
      time_limit_given = true;
    }
    else if (strcmp(argv[iarg],"-maxiter")==0) {
      n_of_iter=atoi(argv[++iarg]);
      iter_limit_given = true;
    }
    else if (strcmp(argv[iarg],"-nants")==0) {
      n_of_ants = atoi(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-lrate")==0) {
      l_rate = atof(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-n")==0) {
      n_of_trials=atoi(argv[++iarg]);
    }
    else if (strcmp(argv[iarg],"-ls")==0) {
      if (strcmp(argv[++iarg],"yes")==0) {
	ls = true;
      }
      else {
	ls = false;
      }
    }

    iarg++;
  }
  if (input_file_given == false) {
    printf("No input file given.");
    printf("\n");
    exit(1);
  }
  if ((time_limit_given == false) && (iter_limit_given == false)) {
    cout << endl;
    cout << "please specify:" << endl;
    cout << endl;
    cout << "* a time limit (i.e., -t 20.0), or" << endl;
    cout << "* an iteration limit (i.e., -maxiter 1000), or" << endl;
    cout << "* both" << endl;
    cout << endl;
    exit(1);
  }
}

/* The method getBestSolutionInIteration can be called after each ant has produced a solution. 
   The method returns the best one among the current solutions produced by the ants */

Solution* getBestSolutionInIteration() {

  Solution* sol;
  bool start = true;
  for (list<Unconstrained_Ant*>::iterator anAnt = ants.begin(); anAnt != ants.end(); anAnt++) {
    if (start) {
      start = false;
      sol = (*anAnt)->constructed_solution;
    }
    else {
      Solution* current = (*anAnt)->constructed_solution;
      if ((*current).quality > (*sol).quality) {
	sol = current;
      }
    }
  }
  return sol->copy();
}

/* The method getIterationAverage can be called after each ant has produced a solution. 
   The method returns the average objective function values of the current solutions produced by the ants */

double getIterationAverage() {

  double ret_val = 0.0;
  for (list<Unconstrained_Ant*>::iterator anAnt = ants.begin(); anAnt != ants.end(); anAnt++) {
    ret_val = ret_val + ((double)((*anAnt)->constructed_solution)->quality);
  }
  return (ret_val / (double)n_of_ants);
}

/* The method resetUniformPheromoneValues initializes (or resets) all the pheromone values to 0.5 */

void resetUniformPheromoneValues() {

  for (int i = 0; i < problem_size; i++) {
    (*pheromone)[i].first = 0.5;
    (*pheromone)[i].second = 0.5;
  }
}

/* The method computeConvergenceFactor computes the convergence factor cf, which gives an
   indication about the current state of the system in terms of its convergence */

double computeConvergenceFactor() {

  double ret_val = 0.0;
  for (int i = 0; i < problem_size; i++) {
    if ((tau_max - (*pheromone)[i].first) > ((*pheromone)[i].first - tau_min)) {
      ret_val = ret_val + tau_max - (*pheromone)[i].first;
    }
    else {
      ret_val = ret_val + (*pheromone)[i].first - tau_min;
    }
    if ((tau_max - (*pheromone)[i].second) > ((*pheromone)[i].second - tau_min)) {
      ret_val = ret_val + tau_max - (*pheromone)[i].second;
    }
    else {
      ret_val = ret_val + (*pheromone)[i].second - tau_min;
    }
  }
  ret_val = ret_val / (2.0 * ((double)problem_size) * (tau_max - tau_min));
  ret_val = (ret_val - 0.5) * 2.0;
  return ret_val;
}

/* 'main' is the main body of the program */

int main( int argc, char **argv ) {

  // first the command line parameters have to be read
  if ( argc < 2 ) {
    exit(1);
  }
  else {
    read_parameters(argc,argv);
  }

  // initialization of the random generator (using the current time)
  rnd = new Random((unsigned) time(&t));

  // initialization of the structure that holds the pheromone values
  pheromone = new map<int,pair<double,double> >;

  /* initialization of the list of ants (as information they receive the problem size,
     the random generator, the pheromone value structure, and the problem instance information */
  for (int na = 0; na < n_of_ants; na++) {
    Unconstrained_Ant* ant = new Unconstrained_Ant(problem_size,rnd,pheromone,q_matrix,c_const,ls);
    ants.push_back(ant);
  }

  /* initialization of the three solutions 'best_so_far', 'restart_best' and 'iteration_best', which
     are used to update the pheromone values */
  Solution* best_so_far = NULL;
  Solution* restart_best = NULL;
  Solution* iteration_best = NULL;

  /* the following three variables are for collecting statistics on several trials */
  int best_solution_value = 0;
  vector<double> results;
  vector<double> times_best_found;

  /* the following for loop is for controlling the number of trials as specified by command line parameters */
  for (int trial_counter = 1; trial_counter <= n_of_trials; trial_counter++) {

    printf("begin try %d\n",trial_counter);

    // upon declaration of a variable of type 'Timer' the time is running ... 
    Timer timer;

    // 'iter' is the iteration counter
    int iter = 1;

    /* if the three solutions that are used for updating the pheromone values are initialized by a 
       previous trial, we delete them */
    if (best_so_far != NULL) {
      delete(best_so_far);
    }
    best_so_far = NULL;
    if (iteration_best != NULL) {
      delete(iteration_best);
    }
    iteration_best = NULL;
    if (restart_best != NULL) {
      delete(restart_best);
    }
    restart_best = NULL;

    // for every trial we reinitialize the pheromone values to 0.5 each */
    resetUniformPheromoneValues();

    /* the following four variable are for controlling the update and the restart of the algorithm 
       cf is the convergence factor
       bs_update regulates the use of the best_so_far solution for updating
       program_stop controls the termination of a trial
       restart controls the restart mechanism of the algorithm 
       time_taken is a variable for keeping the CPU times when checked */
    double cf = 0.0;
    bool bs_update = false;
    bool program_stop = false;
    bool restart = false;
    double time_taken = 0.0;

    /* this is the main loop of the algorithm. At each iteration ants produce a solution each and 
       the pheromone values are updated. */
    while (!program_stop) {

      // every ant constructs a solution, which is then evaluated 
      map<int,pair<double,double> > sgdupd;
      for (list<Unconstrained_Ant*>::iterator anAnt = ants.begin(); anAnt != ants.end(); anAnt++) {
	(*anAnt)->construct();
	(*anAnt)->evaluate();
      }
      
      // the iteration_best solution is determined
      if (iteration_best != NULL) {
	delete(iteration_best);
      }
      iteration_best = getBestSolutionInIteration();
           
      // .. as well as the average objective function values of the solutions in that iteration
      double average = getIterationAverage();

      if (iter == 1) {
	// if we are in the first iteration then we can initialize all the variables
	if (best_so_far != NULL) {
	  delete(best_so_far);
	}
	best_so_far = iteration_best->copy();
	if (restart_best != NULL) {
	  delete(restart_best);
	}
	restart_best = iteration_best->copy();
	time_taken = timer.elapsed_time(Timer::VIRTUAL);
	results.push_back((double)(best_so_far->quality - c_const));
	times_best_found.push_back(time_taken);
	printf("best %g\ttime %f\n",(best_so_far->quality - c_const),time_taken);
      }
      else {
	if (restart) {
	  // if this is the first iteration after a restart, then we do the following:
	  restart = false;
	  delete(restart_best);
	  restart_best = iteration_best->copy();
	  if (iteration_best->quality > best_so_far->quality) {
	    delete(best_so_far);
	    best_so_far = iteration_best->copy();
	    time_taken = timer.elapsed_time(Timer::VIRTUAL);
	    printf("best %g\ttime %f\n",(best_so_far->quality - c_const),time_taken);
	    results[trial_counter-1] = (double)(best_so_far->quality - c_const);
	    times_best_found[trial_counter-1] = time_taken;
	  }
	}
	else {
	  if (iteration_best->quality > restart_best->quality) {
	    delete(restart_best);
	    restart_best = iteration_best->copy();
	  }
	  if (iteration_best->quality > best_so_far->quality) {
	    delete(best_so_far);
	    best_so_far = iteration_best->copy();
	    time_taken = timer.elapsed_time(Timer::VIRTUAL);
	    printf("best %g\ttime %f\n",(best_so_far->quality - c_const),time_taken);
	    results[trial_counter-1] = (double)(best_so_far->quality - c_const);
	    times_best_found[trial_counter-1] = time_taken;
	  }
	}
      }
      if ((best_so_far->quality - c_const) > best_solution_value) {
	best_solution_value = (int)(best_so_far->quality - c_const);
      }
      
      // computation of the convergence factor
      cf = computeConvergenceFactor();

      /* if the best_so_far solution was used for updating the pheromone values and the
         convergence factor is greater than 0.99 we do a restart ... */
      if (bs_update && (cf > 0.99)) {
	bs_update = false;
	restart = true;
	cf = 0.0;
	resetUniformPheromoneValues();
      }
      else {
	/* ... otherwise: if convergence factor is greater than 0.99 we use the best_so_far
           solution from now on for updating */
	if (cf > 0.99) {
	  bs_update = true;
	}
	/* i_weight, r_weight, g_weight are the weights of influence for updating the pheromone values 
           they are set depending on the convergence factor cf */
	double i_weight = 0.0;
	double r_weight = 0.0;
	double g_weight = 0.0;
	if (!bs_update) {
	  if (cf < 0.4) {
	    i_weight = 1.0;
	    r_weight = 0.0;
	    g_weight = 0.0;
	  }
	  if ((cf >= 0.4) && (cf < 0.6)) {
	    i_weight = 2.0 / 3.0;
	    r_weight = 1.0 / 3.0;
	    g_weight = 0.0;
	  }
	  if ((cf >= 0.6) && (cf < 0.8)) {
	    i_weight = 1.0 / 3.0;
	    r_weight = 2.0 / 3.0;
	    g_weight = 0.0;
	  }
	  if (cf >= 0.8) {
	    i_weight = 0.0;
	    r_weight = 1.0;
	    g_weight = 0.0;
	  }
	}
	else {
	  // if bs_update = TRUE we use the best_so_far solution for updating the pheromone values
	  i_weight = 0.0;
	  r_weight = 0.0;
	  g_weight = 1.0;
	}

	/* We specifiy vector d, and then we update the pheromone vector towards vector d
	   depending on the learning rate l_rate */
	map<int,pair<double,double> > d;
	for (int i = 0; i < problem_size; i++) {
	  d[i].first = 0.0;
	  d[i].second = 0.0;
	}
	double sum = (i_weight * iteration_best->quality) + (g_weight * best_so_far->quality) + (r_weight * restart_best->quality);
	for (int i = 0; ((i < problem_size) && (i_weight > 0.0)); i++) {
	  if ((*iteration_best).solution[i] == 0) {
	    d[i].first = d[i].first + (i_weight * iteration_best->quality);
	  }
	  else {
	    d[i].second = d[i].second + (i_weight * iteration_best->quality);
	  }
	}
	for (int i = 0; ((i < problem_size) && (r_weight > 0.0)); i++) {
	  if ((*restart_best).solution[i] == 0) {
	    d[i].first = d[i].first + (r_weight * restart_best->quality);
	  }
	  else {
	    d[i].second = d[i].second + (r_weight * restart_best->quality);
	  }
	}
	for (int i = 0; ((i < problem_size) && (bs_update)); i++) {
	  if ((*best_so_far).solution[i] == 0) {
	    d[i].first = d[i].first + (g_weight * best_so_far->quality);
	  }
	  else {
	    d[i].second = d[i].second + (g_weight * best_so_far->quality);
	  }
	}

	for (int i = 0; i < problem_size; i++) {
	  d[i].first = d[i].first / sum;
	  d[i].second = d[i].second / sum;
	}
	for (int i = 0; i < problem_size; i++) {
	  (*pheromone)[i].first = (*pheromone)[i].first + (l_rate * (d[i].first - (*pheromone)[i].first));
	  (*pheromone)[i].second = (*pheromone)[i].second + (l_rate * (d[i].second - (*pheromone)[i].second));
	  if ((*pheromone)[i].first > tau_max) {
	    (*pheromone)[i].first = tau_max;
	  }
	  if ((*pheromone)[i].second > tau_max) {
	    (*pheromone)[i].second = tau_max;
	  }
	  if ((*pheromone)[i].first < tau_min) {
	    (*pheromone)[i].first = tau_min;
	  }
	  if ((*pheromone)[i].second < tau_min) {
	    (*pheromone)[i].second = tau_min;
	  }
	}
      }
      
      // in the following we check if the termination criteria are satisfied
      iter = iter + 1;
      
      if (time_limit_given && iter_limit_given) {
	if ((timer.elapsed_time(Timer::VIRTUAL) > time_limit) || (iter > n_of_iter)) {
	  program_stop = true;
	}
      }
      else {
	if (time_limit_given) {
	  if (timer.elapsed_time(Timer::VIRTUAL) > time_limit) {
	    program_stop = true;
	  }
	}
	else {
	  if (iter > n_of_iter) {
	    program_stop = true;
	  }
	}
      }
    }
    printf("end try %d\n",trial_counter);
  }

  /* the following lines are for writing the statistics about the experiments onto the screen
     1) the best solution found in all the trials
     2) the average of the best solutions found in all the trials
     3) the standard deviation of the average in 2)
     4) the average CPU at which the best solutions of the trials were found
     5) the standard deviation of the average in 4) */

  double r_mean = 0.0;
  double t_mean = 0.0;
  for (int i = 0; i < results.size(); i++) {
    r_mean = r_mean + results[i];
    t_mean = t_mean + times_best_found[i];
  }
  r_mean = r_mean / ((double)results.size());
  t_mean = t_mean / ((double)times_best_found.size());
  double rsd = 0.0;
  double tsd = 0.0;
  for (int i = 0; i < results.size(); i++) {
    rsd = rsd + pow(results[i]-r_mean,2.0);
    tsd = tsd + pow(times_best_found[i]-t_mean,2.0);
  }
  rsd = rsd / ((double)(results.size()-1.0));
  if (rsd > 0.0) {
    rsd = sqrt(rsd);
  }
  tsd = tsd / ((double)(times_best_found.size()-1.0));
  if (tsd > 0.0) {
    tsd = sqrt(tsd);
  }
  printf("statistics\t%d\t%f\t%f\t%f\t%f\n",best_solution_value,r_mean,rsd,t_mean,tsd);
  delete(rnd);
}
